import { __decorate } from "tslib";
import { $log, Appender, BaseAppender } from "@tsed/logger";
import axios from "axios";
import axiosRetry from "axios-retry";
import * as util from "util";
import { v4 } from "uuid";
function wrapErrorsWithInspect(items) {
    return items.map((item) => {
        if (item instanceof Error && item.stack) {
            return {
                inspect: function () {
                    return `${util.format(item)}\n${item.stack}`;
                }
            };
        }
        return item;
    });
}
function format(logData) {
    return logData ? util.format(...wrapErrorsWithInspect(logData)) : logData;
}
export class LogStashHttpOptions {
}
let LogStashHttpAppender = class LogStashHttpAppender extends BaseAppender {
    #buffer = [];
    #timer;
    build() {
        if ($log.level !== "OFF" && this.config.options) {
            this.client = axios.create({
                baseURL: this.config.options.url,
                auth: this.config.options.auth,
                timeout: this.config.options.timeout || 5000,
                params: this.config.options.params,
                headers: {
                    ...(this.config.options.headers || {})
                },
                httpsAgent: this.config.options.httpsAgent
            });
            axiosRetry(this.client, {
                retries: 3,
                retryDelay: axiosRetry.exponentialDelay,
                ...this.config.options.retryOptions
            });
        }
        if (this.config.options.delayToFlush) {
            this.#timer = setInterval(() => this.flush(), this.config.options.delayToFlush);
        }
    }
    write(loggingEvent) {
        const { logChannel } = this.config.options;
        const level = loggingEvent.level.toString().toLowerCase();
        if (level !== "off") {
            const { time, ...props } = loggingEvent.getData();
            this.send({
                ...props,
                message: format(loggingEvent.getMessage()),
                context: loggingEvent.context.toJSON(),
                level: loggingEvent.level.level / 100,
                level_name: level,
                channel: logChannel,
                datetime: new Date(time || loggingEvent.startTime).toISOString()
            });
        }
    }
    send(bulk) {
        const { bufferMax = 0, delayToFlush = 0 } = this.config.options;
        this.#buffer.push(bulk);
        if ((!bufferMax && !delayToFlush) || (bufferMax && bufferMax <= this.#buffer.length)) {
            return this.flush();
        }
    }
    async flush() {
        // send to server
        const buffer = this.#buffer;
        this.#buffer = [];
        if (!buffer.length) {
            return this;
        }
        const { url, application, logType, requireAlias, debug } = this.config.options;
        const _index = typeof application === "function" ? application() : application;
        const bulkData = buffer.flatMap((item) => [
            JSON.stringify({
                index: {
                    _index,
                    _type: logType,
                    require_alias: requireAlias,
                    _id: v4()
                }
            }),
            item
        ]);
        try {
            const result = await this.client({
                url: "",
                method: "POST",
                data: this.serializeBulk(bulkData),
                headers: {
                    "Content-Type": "application/x-ndjson"
                }
            });
            if (debug) {
                console.debug(`Ts.ED Logger.logstash-http Appender ${url}: ${result.status} - ${JSON.stringify(result.data)}`);
            }
        }
        catch (error) {
            if (error.response) {
                console.error(`Ts.ED Logger.logstash-http Appender error posting to ${url}: ${error.response.status} - ${JSON.stringify(error.response.data)}`);
                return;
            }
            console.error(`Ts.ED Logger.logstash-http Appender error: ${error.message}`);
        }
    }
    serializeBulk(array) {
        return array.reduce((ndjson, obj) => {
            const str = typeof obj === "string" ? obj : JSON.stringify(obj);
            return ndjson + str + "\n";
        }, "");
    }
    shutdown() {
        this.#timer && clearInterval(this.#timer);
        return this.flush();
    }
};
LogStashHttpAppender = __decorate([
    Appender({ name: "logstash-http" })
], LogStashHttpAppender);
export { LogStashHttpAppender };
//# sourceMappingURL=LogStashHttpAppender.js.map